@@ -1,9 +1,7 @@
 /*
- * "$Id: mxml-node.c 462 2016-06-11 20:51:49Z msweet $"
+ * Node support code for Mini-XML, a small XML file parsing library.
  *
- * Node support code for Mini-XML, a small XML-like file parsing library.
- *
- * Copyright 2003-2016 by Michael R Sweet.
+ * Copyright 2003-2017 by Michael R Sweet.
  *
  * These coded instructions, statements, and computer programs are the
  * property of Michael R Sweet and are protected by Federal copyright
@@ -11,7 +9,7 @@
  * which should have been included with this file.  If this file is
  * missing or damaged, see the license at:
  *
- *     http://www.msweet.org/projects.php/Mini-XML
+ *     https://michaelrsweet.github.io/mxml
  */
 
 /*
@@ -33,18 +31,18 @@ static mxml_node_t	*mxml_new(mxml_node_t *parent, mxml_type_t type);
 /*
  * 'mxmlAdd()' - Add a node to a tree.
  *
- * Adds the specified node to the parent. If the child argument is not
- * NULL, puts the new node before or after the specified child depending
- * on the value of the where argument. If the child argument is NULL,
- * puts the new node at the beginning of the child list (MXML_ADD_BEFORE)
- * or at the end of the child list (MXML_ADD_AFTER). The constant
- * MXML_ADD_TO_PARENT can be used to specify a NULL child pointer.
+ * Adds the specified node to the parent.  If the child argument is not
+ * @code NULL@, puts the new node before or after the specified child depending
+ * on the value of the where argument.  If the child argument is @code NULL@,
+ * puts the new node at the beginning of the child list (@code MXML_ADD_BEFORE@)
+ * or at the end of the child list (@code MXML_ADD_AFTER@).  The constant
+ * @code MXML_ADD_TO_PARENT@ can be used to specify a @code NULL@ child pointer.
  */
 
 void
 mxmlAdd(mxml_node_t *parent,		/* I - Parent node */
-        int         where,		/* I - Where to add, MXML_ADD_BEFORE or MXML_ADD_AFTER */
-        mxml_node_t *child,		/* I - Child node for where or MXML_ADD_TO_PARENT */
+        int         where,		/* I - Where to add, @code MXML_ADD_BEFORE@ or @code MXML_ADD_AFTER@ */
+        mxml_node_t *child,		/* I - Child node for where or @code MXML_ADD_TO_PARENT@ */
 	mxml_node_t *node)		/* I - Node to add */
 {
 #ifdef DEBUG
@@ -172,7 +170,7 @@ mxmlAdd(mxml_node_t *parent,		/* I - Parent node */
  * 'mxmlDelete()' - Delete a node and all of its children.
  *
  * If the specified node has a parent, this function first removes the
- * node from its parent using the mxmlRemove() function.
+ * node from its parent using the @link mxmlRemove@ function.
  */
 
 void
@@ -221,29 +219,23 @@ mxmlDelete(mxml_node_t *node)		/* I - Node to delete */
 
     if ((next = current->next) == NULL)
     {
-      mxml_node_t *temp = current->parent;
-					/* Pointer to parent node */
-
-      if (temp == node)
-      {
-       /*
-        * Got back to the top node...
-        */
+     /*
+      * Next node is the parent, which we'll free as needed...
+      */
 
+      if ((next = current->parent) == node)
         next = NULL;
-      }
-      else if ((next = temp->next) == NULL)
-      {
-	if ((next = temp->parent) == node)
-	  next = NULL;
-      }
     }
 
+   /*
+    * Free child...
+    */
+
     mxml_free(current);
   }
 
  /*
-  * Then free the memory used by this node...
+  * Then free the memory used by the parent node...
   */
 
   mxml_free(node);
@@ -282,15 +274,16 @@ mxmlGetRefCount(mxml_node_t *node)	/* I - Node */
  * 'mxmlNewCDATA()' - Create a new CDATA node.
  *
  * The new CDATA node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
- * CDATA node has no parent. The data string must be nul-terminated and
- * is copied into the new node. CDATA nodes use the MXML_ELEMENT type.
+ * list.  The constant @code MXML_NO_PARENT@ can be used to specify that the new
+ * CDATA node has no parent.  The data string must be nul-terminated and
+ * is copied into the new node.  CDATA nodes currently use the
+ * @code MXML_ELEMENT@ type.
  *
  * @since Mini-XML 2.3@
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewCDATA(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewCDATA(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
 	     const char  *data)		/* I - Data string */
 {
   mxml_node_t	*node;			/* New node */
@@ -323,8 +316,8 @@ mxmlNewCDATA(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
  * 'mxmlNewCustom()' - Create a new custom data node.
  *
  * The new custom node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
- * element node has no parent. NULL can be passed when the data in the
+ * list. The constant @code MXML_NO_PARENT@ can be used to specify that the new
+ * element node has no parent. @code NULL@ can be passed when the data in the
  * node is not dynamically allocated or is separately managed.
  *
  * @since Mini-XML 2.1@
@@ -332,7 +325,7 @@ mxmlNewCDATA(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
 
 mxml_node_t *				/* O - New node */
 mxmlNewCustom(
-    mxml_node_t              *parent,	/* I - Parent node or MXML_NO_PARENT */
+    mxml_node_t              *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
     void                     *data,	/* I - Pointer to data */
     mxml_custom_destroy_cb_t destroy)	/* I - Function to destroy data */
 {
@@ -362,12 +355,12 @@ mxmlNewCustom(
  * 'mxmlNewElement()' - Create a new element node.
  *
  * The new element node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
+ * list. The constant @code MXML_NO_PARENT@ can be used to specify that the new
  * element node has no parent.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewElement(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewElement(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
                const char  *name)	/* I - Name of element */
 {
   mxml_node_t	*node;			/* New node */
@@ -400,12 +393,12 @@ mxmlNewElement(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
  * 'mxmlNewInteger()' - Create a new integer node.
  *
  * The new integer node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
+ * list. The constant @code MXML_NO_PARENT@ can be used to specify that the new
  * integer node has no parent.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewInteger(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewInteger(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
                int         integer)	/* I - Integer value */
 {
   mxml_node_t	*node;			/* New node */
@@ -429,14 +422,14 @@ mxmlNewInteger(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
 /*
  * 'mxmlNewOpaque()' - Create a new opaque string.
  *
- * The new opaque node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
- * opaque node has no parent. The opaque string must be nul-terminated and
- * is copied into the new node.
+ * The new opaque string node is added to the end of the specified parent's
+ * child list.  The constant @code MXML_NO_PARENT@ can be used to specify that
+ * the new opaque string node has no parent.  The opaque string must be nul-
+ * terminated and is copied into the new node.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewOpaque(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewOpaque(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
               const char  *opaque)	/* I - Opaque string */
 {
   mxml_node_t	*node;			/* New node */
@@ -465,16 +458,62 @@ mxmlNewOpaque(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
 }
 
 
+/*
+ * 'mxmlNewOpaquef()' - Create a new formatted opaque string node.
+ *
+ * The new opaque string node is added to the end of the specified parent's
+ * child list.  The constant @code MXML_NO_PARENT@ can be used to specify that
+ * the new opaque string node has no parent.  The format string must be
+ * nul-terminated and is formatted into the new node.
+ */
+
+mxml_node_t *				/* O - New node */
+mxmlNewOpaquef(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
+               const char  *format,	/* I - Printf-style format string */
+	       ...)			/* I - Additional args as needed */
+{
+  mxml_node_t	*node;			/* New node */
+  va_list	ap;			/* Pointer to arguments */
+
+
+#ifdef DEBUG
+  fprintf(stderr, "mxmlNewOpaquef(parent=%p, format=\"%s\", ...)\n", parent, format ? format : "(null)");
+#endif /* DEBUG */
+
+ /*
+  * Range check input...
+  */
+
+  if (!format)
+    return (NULL);
+
+ /*
+  * Create the node and set the text value...
+  */
+
+  if ((node = mxml_new(parent, MXML_OPAQUE)) != NULL)
+  {
+    va_start(ap, format);
+
+    node->value.opaque = _mxml_vstrdupf(format, ap);
+
+    va_end(ap);
+  }
+
+  return (node);
+}
+
+
 /*
  * 'mxmlNewReal()' - Create a new real number node.
  *
  * The new real number node is added to the end of the specified parent's
- * child list. The constant MXML_NO_PARENT can be used to specify that
+ * child list.  The constant @code MXML_NO_PARENT@ can be used to specify that
  * the new real number node has no parent.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewReal(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewReal(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
             double      real)		/* I - Real number value */
 {
   mxml_node_t	*node;			/* New node */
@@ -499,14 +538,14 @@ mxmlNewReal(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
  * 'mxmlNewText()' - Create a new text fragment node.
  *
  * The new text node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
- * text node has no parent. The whitespace parameter is used to specify
- * whether leading whitespace is present before the node. The text
+ * list.  The constant @code MXML_NO_PARENT@ can be used to specify that the new
+ * text node has no parent.  The whitespace parameter is used to specify
+ * whether leading whitespace is present before the node.  The text
  * string must be nul-terminated and is copied into the new node.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewText(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewText(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
             int         whitespace,	/* I - 1 = leading whitespace, 0 = no whitespace */
 	    const char  *string)	/* I - String */
 {
@@ -543,16 +582,16 @@ mxmlNewText(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
  * 'mxmlNewTextf()' - Create a new formatted text fragment node.
  *
  * The new text node is added to the end of the specified parent's child
- * list. The constant MXML_NO_PARENT can be used to specify that the new
- * text node has no parent. The whitespace parameter is used to specify
- * whether leading whitespace is present before the node. The format
+ * list.  The constant @code MXML_NO_PARENT@ can be used to specify that the new
+ * text node has no parent.  The whitespace parameter is used to specify
+ * whether leading whitespace is present before the node.  The format
  * string must be nul-terminated and is formatted into the new node.
  */
 
 mxml_node_t *				/* O - New node */
-mxmlNewTextf(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
+mxmlNewTextf(mxml_node_t *parent,	/* I - Parent node or @code MXML_NO_PARENT@ */
              int         whitespace,	/* I - 1 = leading whitespace, 0 = no whitespace */
-	     const char  *format,	/* I - Printf-style frmat string */
+	     const char  *format,	/* I - Printf-style format string */
 	     ...)			/* I - Additional args as needed */
 {
   mxml_node_t	*node;			/* New node */
@@ -592,8 +631,8 @@ mxmlNewTextf(mxml_node_t *parent,	/* I - Parent node or MXML_NO_PARENT */
 /*
  * 'mxmlRemove()' - Remove a node from its parent.
  *
- * Does not free memory used by the node - use mxmlDelete() for that.
- * This function does nothing if the node has no parent.
+ * This function does not free memory used by the node - use @link mxmlDelete@
+ * for that.  This function does nothing if the node has no parent.
  */
 
 void
@@ -660,7 +699,7 @@ mxmlRemove(mxml_node_t *node)		/* I - Node to remove */
  * 'mxmlNewXML()' - Create a new XML document tree.
  *
  * The "version" argument specifies the version number to put in the
- * ?xml element node. If NULL, version 1.0 is assumed.
+ * ?xml element node. If @code NULL@, version "1.0" is assumed.
  *
  * @since Mini-XML 2.3@
  */
@@ -682,7 +721,7 @@ mxmlNewXML(const char *version)		/* I - Version number to use */
  * 'mxmlRelease()' - Release a node.
  *
  * When the reference count reaches zero, the node (and any children)
- * is deleted via mxmlDelete().
+ * is deleted via @link mxmlDelete@.
  *
  * @since Mini-XML 2.3@
  */
@@ -835,8 +874,3 @@ mxml_new(mxml_node_t *parent,		/* I - Parent node */
 
   return (node);
 }
-
-
-/*
- * End of "$Id: mxml-node.c 462 2016-06-11 20:51:49Z msweet $".
- */
